open Lexing

module SMap = Map.Make(String)
module SSet = Set.Make(String)

type pvar = string
type pchannel  = string

type var = int
type channel = string

type pprog =
  | PSend of pchannel * pprog
  | PRec  of pchannel * pvar * pprog
  | PComp of pprog * pprog
  | PExe  of pvar
  | PZero

type disp_prog = DProg of (((disp_send list) * (disp_receive list)) SMap.t
                           * (pvar list))
and disp_receive = pvar * disp_prog
and disp_send = disp_prog

type prog   = Prog of (inprog * (var list))
and send    = prog
and receive = prog
and inprog  = ((send list) * (receive list)) SMap.t

type 'a annot = 'a * env
and env = Env of prog annot list
type intop = (send annot) list * (receive annot) list
type toplevel = intop SMap.t

let ($) = List.rev_append

let channel_empty = SMap.empty
let channel_fold  = SMap.fold
let channel_iter  = SMap.iter
let channel_size  = SMap.cardinal
let channel_map   = SMap.map

let in_chan_fold  = List.fold_left

let var_empty = []
let var_iter  = List.iter
let var_is_empty  x = x = []
let var_singleton x = [x]
let var_union = List.rev_append
let var_fold  = List.fold_left
let var_size  = List.length

let rec rev_concat l' = function
  | a::l -> rev_concat (a $ l') l
  | []   -> l'



