open Ast
open Utils

let incr_env m =
  SMap.map (fun x -> x + 1) m

exception Not_closed of string

(** pestle_prog transforme un pprog en prog en allant chercher dans
 * l'environnement la valeur à affecter aux variables représentées par leur
 * indice de de Bruijn
*)
let rec pestle_prog env = function
  | PSend (a,p)  -> Prog (SMap.singleton a ([pestle_prog env p], []), [])
  | PRec (a,x,p)
    -> Prog (SMap.singleton a
               ([], [pestle_prog (SMap.add x 0 (incr_env env)) p]), [])
  | PComp (p,q)  -> merge (pestle_prog env p) (pestle_prog env q)
  | PExe x -> Prog (empty, [try SMap.find x env
                            with Not_found -> raise (Not_closed x)])
  | PZero  -> Prog (empty, [])

(** pestle transforme un pprog en toplevel. Le terme doit être clos *)
let rec pestle = function
  | PZero  -> (SMap.empty: toplevel)
  | PSend (a,p)  -> SMap.singleton a ([pestle_prog empty p, Env []], [])
  | PRec (a,x,p)
    -> SMap.singleton a ([], [pestle_prog (SMap.singleton x 0) p, Env []])
  | PComp (p,q)  -> let p = pestle p in
    let q = pestle q in shrink p q
  | PExe x -> assert false

