open Ast
open Utils
open Printer
open Format
open Scanf


exception Stop
exception Backtrack


let ask message max f =
  let rec aux f =
    printf "%s @?" message;
    try bscanf Scanning.stdin "%u\n" (fun x ->
        if x >= max
        then (printf "Invalid value: out of bounds@."; aux f)
        else f x)
    with
    | Scanf.Scan_failure _ | Failure _ | End_of_file | Invalid_argument _ ->
      raise Stop in
  aux f


let oracle_ask print_send print_recv empty (ls, lr) =
  let print_annot_send n (x,_) =
  printf "\tn°%d: @[%a@]@." n print_send x;
  n + 1 in
  let print_annot_recv n (x,_) =
  printf "\tn°%d: @[%a@]@." n print_recv x;
  n + 1 in
  let propose to_print l =
    List.fold_left (fun n (a,l) ->
        printf " - On channel %s:@." a;
        List.fold_left to_print n l) 1 l in
  printf "Emissions: @.";
  let max_send = propose print_annot_send ls in
  printf "@.Receptions: @.";
  let max_rec  = propose print_annot_recv lr in
  printf "@.";

  let find_send n l =
    let r = ref ("",empty) in
    let aux a =
      List.fold_left (fun k x ->
          if k = n then (r := (a,x); raise Stop) else k + 1) in
    try ignore (List.fold_left (fun k (a,l) -> aux a k l) 1 l);
      raise Not_found
    with Stop -> !r in

  let find_rec n l =
    let r = ref ("",empty) in
    let aux a =
      List.fold_left (fun k x ->
          if k = n then raise (r := (a,x); raise Stop) else k + 1) in
    try ignore (List.fold_left (fun k (a,l) -> aux a k l) 1 l);
      raise Not_found
    with Stop -> !r in

  let rec ask_valid () =
    ask "Choose an emission (0 to backtrack):" max_send (fun i_send ->
        if i_send = 0 then raise Backtrack;
        ask "Choose a reception (0 to backtrack):" max_rec (fun i_rec ->
            if i_rec = 0 then raise Backtrack;
            let (a,s) = find_send i_send ls in
            let (b,r) = find_rec  i_rec  lr in
            if a <> b then (printf "Not the same channel@.@."; ask_valid ())
            else (a, s, r))) in
  ask_valid ()


let machine_oracle print_top oracle m =
  printf "@. - - - - - - -@.Current state: @.@.%a@.@." print_top m;
  let make_choices a (s,r) ((ls, lr) as l) =
    if s = [] || r = [] then l else
      ((a, prune s)::ls, (a, prune r)::lr) in
  let ((ls, lr) as l) = channel_fold make_choices m ([],[]) in
  if ls = [] || lr = [] then (printf "No reductions left@."; raise Stop);
  let (a, (s, s_list), ((r, Env r_env), r_list)) = oracle l in
  let to_shrink = eval r (Env (s::r_env)) in
  printf "New process(es) to add: %a@." print_top to_shrink;
  (shrink to_shrink (SMap.add a (s_list, r_list) m): toplevel)


let rec machine_ask m =
  let annot_empty : prog annot = (Prog (empty, []), Env []) in
  let print_obj = pretty_print_prog_annot in
  let oracle = oracle_ask print_obj print_obj  (annot_empty, []) in
  let rec to_call m father =
    try
      to_call (machine_oracle pretty_print_real_top oracle m)
        (fun _ -> to_call m father)
    with
    | Stop -> printf "Final state reached:@.@.%a@.@." pretty_print_top m;
      ask "Bactrack (0) ? Stop (1) ?" 2 (fun x -> if x = 0 then father ())
    | Backtrack -> father () in
  to_call m (fun _ -> printf "Cannot backtrack further@."; machine_ask m)
